package com.dzeio.openhealth.extensions

import android.Manifest
import android.app.Activity
import android.content.Intent
import android.content.pm.PackageManager
import android.os.Build
import android.util.Log
import androidx.core.app.ActivityCompat
import com.dzeio.openhealth.data.weight.Weight
import com.google.android.gms.auth.api.signin.GoogleSignIn
import com.google.android.gms.fitness.Fitness
import com.google.android.gms.fitness.FitnessOptions
import com.google.android.gms.fitness.data.DataPoint
import com.google.android.gms.fitness.data.DataType
import com.google.android.gms.fitness.request.DataReadRequest
import java.text.DateFormat
import java.util.*
import java.util.concurrent.TimeUnit

class GoogleFit(
    private val activity: Activity,
) : Extension() {
    companion object {
        const val TAG = "GoogleFitConnector"
    }

    override val sourceID: String = "GoogleFit"

    private val fitnessOptions = FitnessOptions.builder()
        .addDataType(DataType.TYPE_WEIGHT)
        .addDataType(DataType.TYPE_STEP_COUNT_CUMULATIVE)
        .build()

    private fun checkPermissionsAndRun(data: Data) {
        if (permissionApproved()) {
            signIn(data)
        } else {
            Log.d(TAG, "Asking for permission")
            // Ask for permission
            ActivityCompat.requestPermissions(
                activity,
                arrayOf(Manifest.permission.ACCESS_FINE_LOCATION),
                data.ordinal
            )
        }
    }

    private fun permissionApproved(): Boolean {
        val approved = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.Q) {
            PackageManager.PERMISSION_GRANTED == ActivityCompat.checkSelfPermission(
                activity,
                Manifest.permission.ACCESS_FINE_LOCATION)
        } else {
            true
        }
        return approved
    }

    private fun signIn(data: Data) {
        if (oAuthPermissionsApproved()) {
            startImport(data)
        } else {
            Log.d("GoogleFitImporter", "Signing In")
            GoogleSignIn.requestPermissions(
                activity,
                data.ordinal,
                getGoogleAccount(), fitnessOptions)
        }
    }

    private fun oAuthPermissionsApproved() = GoogleSignIn.hasPermissions(getGoogleAccount(), fitnessOptions)

    private fun getGoogleAccount() = GoogleSignIn.getAccountForExtension(activity, fitnessOptions)

    private val timeRange by lazy {
        val calendar = Calendar.getInstance(TimeZone.getTimeZone("UTC"))
        calendar.time = Date()
        val endTime = calendar.timeInMillis

        // Set year to 2013 to be sure to get data from when Google Fit Started to today
        calendar.set(Calendar.YEAR, 2013)
        val startTime = calendar.timeInMillis
        arrayOf(startTime, endTime)
    }


    private fun startImport(data: Data) {
        Log.d("GoogleFitImporter", "Importing for ${data.name}")

        val dateFormat = DateFormat.getDateInstance()
        Log.i(TAG, "Range Start: ${dateFormat.format(timeRange[0])}")
        Log.i(TAG, "Range End: ${dateFormat.format(timeRange[1])}")

        var type = DataType.TYPE_WEIGHT
        var timeUnit = TimeUnit.MILLISECONDS

        when (data) {
            Data.STEPS -> {
                type = DataType.TYPE_STEP_COUNT_CUMULATIVE
            }
            else -> {}
        }

        runRequest(DataReadRequest.Builder()
            .read(type)
            .setTimeRange(timeRange[0], timeRange[1], timeUnit)
            .build(), data)

    }

    private fun runRequest(request: DataReadRequest, data: Data) {
        Fitness.getHistoryClient(activity, GoogleSignIn.getAccountForExtension(activity, fitnessOptions))
            .readData(request)
            .addOnSuccessListener { response ->
                Log.d(TAG, "Received response! ${response.dataSets.size} ${response.buckets.size} ${response.status}")
                for (dataSet in response.dataSets) {
                    Log.i(TAG, "Data returned for Data type: ${dataSet.dataType.name} ${dataSet.dataPoints.size}")
                    dataSet.dataPoints.forEachIndexed { index, dp ->
                        val isLast = (index + 1) == dataSet.dataPoints.size

                        // Global
                        Log.i(TAG,"Importing Data point:")
                        Log.i(TAG,"\tType: ${dp.dataType.name}")
                        Log.i(TAG,"\tStart: ${dp.getStartTimeString()}")
                        Log.i(TAG,"\tEnd: ${dp.getEndTimeString()}")

                        // Field Specifics
                        for (field in dp.dataType.fields) {
                            Log.i(TAG,"\tField: ${field.name} Value: ${dp.getValue(field)}")
                            when (data) {
                                Data.WEIGHT -> {
                                    val weight = Weight()
                                    weight.timestamp = dp.getStartTime(TimeUnit.MILLISECONDS)
                                    weight.weight = dp.getValue(field).asFloat()
                                    weightCallback(weight, isLast)
                                }
                                else -> {}
                            }
                        }
                    }
                }
            }
            .addOnFailureListener { e ->
                Log.e(TAG,"There was an error reading data from Google Fit", e)
            }
    }

    private fun DataPoint.getStartTimeString(): String = Date(this.getStartTime(TimeUnit.SECONDS) * 1000L).toLocaleString()

    private fun DataPoint.getEndTimeString(): String = Date(this.getEndTime(TimeUnit.SECONDS) * 1000L).toLocaleString()

    override fun onRequestPermissionResult(
        requestCode: Int,
        permission: Array<String>,
        grantResult: IntArray
    ) {
        signIn(Data.values()[requestCode])
    }

    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?) {
        signIn(Data.values()[requestCode])
    }

    private lateinit var weightCallback: (weight: Weight, end: Boolean) -> Unit

    override fun importWeight(callback: (weight: Weight, end: Boolean) -> Unit) {
        this.weightCallback = callback
        checkPermissionsAndRun(Data.WEIGHT)
    }

    private lateinit var callback : (item: Any, end: Boolean) -> Unit

    override fun <T> import(data: Data, cb: (item: T, end: Boolean) -> Unit) {
        callback = cb as (item: Any, end: Boolean) -> Unit
        when (data) {
            Data.WEIGHT -> {
                checkPermissionsAndRun(data)
            }
            else -> {
                Log.d(TAG, "PRRRRRRRRRRRRR")
            }
        }
    }

}